module wei_mig_base
  !! WEI migration

  use wei_util
  use wei_run
   
  use wei_obj_S
  use wei_obj_R
  use wei_obj_D
  use wei_obj_W

  use wei_fft
  use wei_kmap
  use wei_xmap

  use wei_kxmig
  use wei_image
  !---------------
  implicit none

  type(wedata) :: D,bW
  type(weimag) :: R
  complex, allocatable,target :: buf(:,:,:,:,:)


contains
!!$
!!$=head1 
!!$
!!$ weimig_init - Initialize migration operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$  call weimig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item SLin - Slowness operator init  L<weislo1_init>
!!$
!!$=item WCin - WCop operator init  L<weikxmig1_init>
!!$
!!$=item FKin - FK operator init  L<weiwem_init>
!!$
!!$=item FXin - FX operator init  L<weissf_init>
!!$
!!$=item IGin - image gather creation operator init  L<weihcig_init>
!!$
!!$=item in6,..in9 -  integer,optional
!!$
!!$ Dummy operators for future expansion
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize migration/modeling operator
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weimig> L<wei>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  subroutine weimig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
    integer,external,optional ::                     in6,in7,in8,in9
    integer :: st,er,ierr
    character(len=128) :: name
    logical :: junkl

#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
       integer function    IGin() result(st)
       end function IGin
    end interface !!---------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif



    call show_mem("first")

    !! initialize the main WEI variables
    if(.not. weirun_init()) &
      call erexit("trouble initializing weirun_init")

#ifdef SEP_TIMER
     if(run%m_ith <2) call start_timer_num(timer%init_t)
#endif

    call show_mem("run")
    if(.not. down_init()) &
      call erexit("trouble initializing down_init")
    call show_mem("down_init")


    !! init K-map
    call weikmap_init()
    call weixmap_init()
    call show_mem("map")


    !! initialize objects
    if(.not. wei_slow_init(S,down%VS,down%VG))&
      call erexit("trouble initializing slowness")
    call show_mem("slow_init")
    if(.not. init_distrib_pattern()) &
      call erexit("trouble initializing distribution pattern")
    call show_mem("distrib_init")
    if(.not. wei_image_init(R,down%R)) &
      call erexit("trouble initializing image")
    call show_mem("wei_image_init")
    if(.not. wei_data_init(D,down%D)) &
      call erexit("trouble initializing data")
    call show_mem("data_init")


    !! auxiliary objects
    name="bwfld.WEI"  
    if(.not. wei_aux_data_init(bW,name))&   !CREATE AN AUXILARY WAVEFIELD FILE
      call erexit("trouble initializing wavefield object")

    if(associated(D%bin)) then
      deallocate(D%bin)
      allocate(buf(down%dsep%n(1),down%dsep%n(2),down%dsep%n(3),down%dsep%n(4),down%nws))
      D%bin=>buf
    end if
    bw%bin=>D%bin
    
    call weifft_init(bw%bin)  
    
    call show_mem("data_init")

    down%mod_ = down%R   !MODEL IS REFLECTIVITY
    down%dat_ = down%D   !DATA IS FREQUENCY SLICES

    !! convert frequency to radians
    call process_radians()

    call allocate_tmps()
    call show_mem("tmp_alloc")



    !! init imaging
    call weiimage_init()
    call show_mem("weiimage_init")

    !! allocate slowness arrays for FK/FX ops
    call weislow_allocate()
    call show_mem("slow_init")

    !! check if offset is consistent with ph

    !! init operators
    st=SLin()     !! FK slownesses init
    call show_mem("sl_in")
    st=WCin()  !! WCop init
    call show_mem("wcin")
    st=FKin()     !! FK op init
    call show_mem("fkin")
    st=FXin()     !! FX op init
    call show_mem("fxin")
    st=IGin()     !! FX op init
    call show_mem("igin")
    !! init FFT
    call show_mem("fft_init")
#ifdef SEP_TIMER
      if(run%m_ith <2) call stop_timer_num(timer%init_t)
#endif

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weimig_init
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weimig_outnull()
  !ZERO OUTPUT IF REQUESTED
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_outnull'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! R<-D
        if(run%m_ith==0) write(0,*) "calling image assign(mig_outnull)"
       call     image_assign(R,(0.,0.))
    else             !! R->D
       call      data_assign(D,(0.,0.))
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weimig_outnull
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  logical function weimig_begin(adj,add,mod_,dat_)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: mod_,dat_
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_begin'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif
    weimig_begin=F

#ifdef SEP_TIMER
    if(run%m_ith==0) call start_timer_num(timer%distrib_t)
#endif
    !! reset tag variables
    call down_reset(adj,add,mod_,dat_)
    
    !! accomodate tag name changes
    R%tag = trim(down%mod_)
    D%tag = trim(down%dat_)
#ifdef SEP_MPI
    if(.not.mpi_image_setup(R)) then
      write(0,*) "trouble setting up image"
      return
    end if
    call show_mem("image_setup")
    if(.not.mpi_data_setup(D)) then
      write(0,*) "trouble setting up data"
      return
    end if
    call show_mem("data_setup")
#endif

    !! setup the adjoint sign
    if(down%adj) then
       down%opsign = goT2B
    else
       down%opsign = goB2T
    end if

   if(.not. down%add .and. run%m_nth==1) call weimig_outnull()
    call init_progress_count()
    call show_mem("progress_count")
!   call seperr("")

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif


#ifdef SEP_TIMER
    if(run%m_ith==0) call stop_timer_num(timer%distrib_t)
#endif
    weimig_begin=T
end function


  logical function weimig_end()
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_end'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
      integer :: ierr
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif
    weimig_end=F

#ifdef SEP_TIMER
    if(run%m_ith==0) call start_timer_num(timer%collect_t)
#endif

#ifdef SEP_MPI
    if(run%verb_all) write(0,*) "thread=",run%m_ith," finished"
    
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
    if(run%verb) call weimessage("COMBINE DATA")
    call combine_data(D)

    if(run%verb) call weimessage("COMBINE IMAGE")
    call combine_image(R)

    if(run%verb) call weimessage("COMBINE SLOW")
!    call combine_slow(S)

    if(run%verb) call weimessage("OK")

    if(run%unlink) call aux_unlink(trim(bW%tag))
if(down%order(1:3) == "VAR" .and. run%unlink) call aux_unlink(S%s_s%tag_var);
    if(ruN%unlink) call aux_unlink(S%s_s%tag)
    if(down%conv_wave .and. run%unlink) call aux_unlink(S%s_g%tag)
#endif
!if(run%m_ith==0 .and. run%unlink) call aux_unlink(bw%tag_0)

#ifdef SEP_TIMER
    if(run%m_ith==0) call stop_timer_num(timer%collect_t)
#ifdef SEP_MPI
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
#endif
    if(run%m_ith==0) call print_timers() 
#ifdef SEP_MPI
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
#endif
    if(run%m_ith==1) then
      write(0,*) "SLAVE THREAD 1"
      call print_timers()
    end if
#endif

   D%from="history"
   R%from="history"
    weimig_end=T
 write(0,*) "EXITING MIG",run%m_ith
end function

  !---------------------------------------------------------------
  !! don't write the wavefield
  subroutine wmigRD(izb,iwb,nzs,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izs,iws,ith,nzs,ierr,i
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigRD'
    if(run%debug) call in(unit)
#endif


    if(down%adj) then !! MIGRATION
if(izb==1) then
  do iws=1,down%nws 
     bw%bin(:,:,:,:,iws)=bw%bin(:,:,:,:,iws)*down%wt_local(iws)
  end do
end if

#ifdef SEP_OMP
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
#endif
       wsA: do iws=1,down%nws             !LOOP OVER w STEPS
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
          ith=1
#endif
          zsA: do izs=1,nzs          !LOOP OVER z STEPS
             call weiimgW2R(izs,iws,ith,R%bin,bW%bin,IGop)
             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)
          end do zsA
       end do wsA
#ifdef SEP_OMP
       !$OMP END DO
       !$OMP END PARALLEL
#endif
    else         !! MODELLING
#ifdef SEP_OMP
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
#endif
       wsF: do iws=1,down%nws             !LOOP OVER w STEPS
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
ith=1
#endif
          zsF: do izs=down%nzs,1,-1       !LOOP OVER z STEPS
             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)
             call weiimgR2W(izs,iws,ith, R%bin,bW%bin,IGop)

          end do zsF
       end do wsF
#ifdef SEP_OMP
       !$OMP END DO
       !$OMP END PARALLEL
#endif
    end if


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigRD
  !----------------------------------------------------------
end module wei_mig_base
