module wei_ssf
  !! Split-Step Fourier

  use wei_util
  use wei_run
  use wei_obj_S
  use wei_down
  !---------------
  implicit none
  complex, pointer, private :: w_extrap_1(:,:,:,:,:),w_extrap_2(:,:,:,:,:) 

contains
  !----------------------------------------------------------------
!!$
!!$=head1 NAME
!!$
!!$ weissf_init - Initialize split step operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weissf_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize split step operator
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weissf>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  integer function weissf_gsp_init() result(st)
    integer :: er,n
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf_init'
    if(run%debug) call in(unit)
#endif
     n=size(tmp_on,5)
     if(run%fk_extra .ne. 2) call seperr("expecting  fk_extra ==2 for gsp")
     w_extrap_1=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     n=n-run%o_nth
     w_extrap_2=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
    !! do nothing
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=weissf_init()
  end function weissf_gsp_init
  !----------------------------------------------------------------
  integer function weissf_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf_init'
    if(run%debug) call in(unit)
#endif
    !! do nothing
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weissf_init
  !----------------------------------------------------------------
!!$
!!$=head1 NAME
!!$
!!$ weissf - Split step operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weissf_init(iws,izs,ifk,ith,wfld)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  Current frequency slice
!!$
!!$=item izs - integer
!!$
!!$  Current depth slice
!!$
!!$=item ifk - integer
!!$
!!$  Current fk operation
!!$
!!$=item ith - integer
!!$
!!$  Current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$ Wavefield
!!$
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Split step operator. Apply first order slowness correction
!!$  to wthe wavefield given difference between velocity at
!!$  location and reference velocity.
!!$
!!$=head1 SEE ALSO
!!!$
!!$L<weissf_init>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
!! Split-step
  integer function weissf_gsp(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) :: iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    ::  w,j,k,ds,ds2
    integer :: i1,i2,i3,i4,n1,n2,n3,n4,nc
    complex :: iwdz
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf'
    if(run%depar) call in(unit)
#endif

#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%fx_t)
#endif



!    w    = down%bw%o + (iws-1)*down%bw%d
    w= down%w_local(iws)
       j =  down%opsign*w*var%dzb(izs)
    iwdz = (0,1)* down%opsign*w*var%dzb(izs)


    n4=size(wfld,4)
    n3=size(wfld,3)
    n2=size(wfld,2)
    n1=size(wfld,1)
    if(run%mem_save<1) then
      if(slo%th(ith)%source) then
        do i4=1,n4;  do i3=1,n3
         do i2=1,n2; do i1=1,n1
           ds=( real(S%pba(i1,i2,i3,i4,izs))-slo%sz_s(izs,ifk))
           ds2= real(S%pba(i1,i2,i3,i4,izs))**2-slo%sz_s(izs,ifk)**2
           k=j*ds
          wfld(i1,i2,i3,i4,iws) =wfld(i1,i2,i3,i4,iws) *cmplx(cos(k),sin(k))
          w_extrap_1(i1,i2,i3,i4,ith)=ds*wfld(i1,i2,i3,i4,iws)
          w_extrap_2(i1,i2,i3,i4,ith)=ds2*ds2*wfld(i1,i2,i3,i4,iws)
         end do; end do; end do; end do;
        else
        do i4=1,n4;  do i3=1,n3
         do i2=1,n2; do i1=1,n1
          ds=( aimag(S%pba(i1,i2,i3,i4,izs))-slo%sz_r(izs,ifk))
          k=j*ds
          ds2= aimag(S%pba(i1,i2,i3,i4,izs))**2-slo%sz_r(izs,ifk)**2
          wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
          w_extrap_1(i1,i2,i3,i4,ith)=ds2*wfld(i1,i2,i3,i4,iws)
          w_extrap_2(i1,i2,i3,i4,ith)=ds2*ds2*wfld(i1,i2,i3,i4,iws)
         end do; end do; end do; end do;
      end if
    else
      if(slo%th(ith)%source) then
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4,ith,izs)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            ds=( (S%pba_s(i1,i2))-slo%sz_s(izs,ifk))
            ds2= (S%pba_s(i1,i2))**2-slo%sz_s(izs,ifk)**2
            k=j*ds
            wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
            w_extrap_1(i1,i2,i3,i4,ith)=ds2*wfld(i1,i2,i3,i4,iws)
            w_extrap_2(i1,i2,i3,i4,ith)=ds2*ds2*wfld(i1,i2,i3,i4,iws)
          end do; end do; end do; end do;
       else
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4,ith,izs)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            ds=( (S%pba_s(i1,i2))-slo%sz_r(izs,ifk))
            ds2= (S%pba_s(i1,i2))**2-slo%sz_r(izs,ifk)**2
            k=j*ds
            wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
            w_extrap_1(i1,i2,i3,i4,ith)=ds2*wfld(i1,i2,i3,i4,iws)
            w_extrap_2(i1,i2,i3,i4,ith)=ds2*ds2*wfld(i1,i2,i3,i4,iws)
          end do; end do; end do; end do;
       end if
    end if

#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fx_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weissf_gsp
  !----------------------------------------------------------------





  integer function weissf(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) :: iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    ::  w,j,k
    integer :: i1,i2,i3,i4,n1,n2,n3,n4
    complex :: iwdz
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf'
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%fx_t)
#endif

!    w    = down%bw%o + (iws-1)*down%bw%d
    w= down%w_local(iws)
       j =  down%opsign*w*var%dzb(izs)
    iwdz = (0,1)* down%opsign*w*var%dzb(izs)


    n4=size(wfld,4)
    n3=size(wfld,3)
    n2=size(wfld,2)
    n1=size(wfld,1)
    if(run%mem_save<1) then
      if(slo%th(ith)%source) then
        do i4=1,n4;  do i3=1,n3
         do i2=1,n2; do i1=1,n1
           k=j*( real(S%pba(i1,i2,i3,i4,izs))-slo%sz_s(izs,ifk))
          wfld(i1,i2,i3,i4,iws) =wfld(i1,i2,i3,i4,iws) *cmplx(cos(k),sin(k))
         end do; end do; end do; end do;
        else
        do i4=1,n4;  do i3=1,n3
         do i2=1,n2; do i1=1,n1
          k=j*( aimag(S%pba(i1,i2,i3,i4,izs))-slo%sz_r(izs,ifk))
          wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
         end do; end do; end do; end do;
      end if
    else
      if(slo%th(ith)%source) then
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4,ith,izs)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            k=j*( (S%pba_s(i1,i2))-slo%sz_s(izs,ifk))
            wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
          end do; end do; end do; end do;
       else
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4,ith,izs)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            k=j*( (S%pba_s(i1,i2))-slo%sz_r(izs,ifk))
            wfld(i1,i2,i3,i4,iws) = wfld(i1,i2,i3,i4,iws)*cmplx(cos(k),sin(k))
          end do; end do; end do; end do;
       end if
    end if

#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fx_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weissf
  !----------------------------------------------------------------
end module wei_ssf
